use crate::AppState;
use actix_web::web::Path;
use actix_web::{
    error::{ErrorBadRequest, ErrorInternalServerError},
    web, Responder,
};
use chrono::Utc;
use entity::{node, node_group, sensor_data};
use eui48::ParseError;
use sea_orm::{entity::*, query::*, ActiveModelTrait, ActiveValue, EntityTrait};
use serde::{Deserialize, Serialize};

#[derive(Serialize)]
struct NodeWithSensorData {
    node: NodeWithMac,
    sensor_data: Vec<sensor_data::Model>,
}

#[derive(Deserialize)]
pub struct CreateGroupWithoutId {
    name: String,
}

#[derive(Deserialize, Serialize)]
pub struct NodeWithMac {
    id: String,
    coord_la: f64,
    coord_lo: f64,
    battery_minimum: f64,
    battery_maximum: f64,
    group: uuid::Uuid,
}

#[derive(Deserialize, Serialize)]
pub struct UpdateNode {
    coord_la: f64,
    coord_lo: f64,
    battery_minimum: f64,
    battery_maximum: f64,
    group: uuid::Uuid,
}

impl From<node::Model> for NodeWithMac {
    fn from(value: node::Model) -> Self {
        let mac = convert_id_to_mac(&value.id);
        Self {
            id: mac,
            coord_la: value.coord_la,
            coord_lo: value.coord_lo,
            battery_minimum: value.battery_minimum,
            battery_maximum: value.battery_maximum,
            group: value.group,
        }
    }
}

impl TryInto<node::Model> for NodeWithMac {
    type Error = eui48::ParseError;

    fn try_into(self) -> Result<node::Model, Self::Error> {
        let mac = convert_mac_to_id(&self.id)?;
        Ok(node::Model {
            id: mac,
            coord_la: self.coord_la,
            coord_lo: self.coord_lo,
            battery_minimum: self.battery_minimum,
            battery_maximum: self.battery_maximum,
            group: self.group,
        })
    }
}

pub fn convert_mac_to_id(mac: &str) -> Result<i64, ParseError> {
    let mac = eui48::MacAddress::parse_str(mac)?;
    let mac_bytes = mac.to_array();
    let mut mac_id_bytes: [u8; 8] = [0; 8];
    mac_id_bytes[2..].copy_from_slice(&mac_bytes);
    Ok(i64::from_be_bytes(mac_id_bytes))
}

pub fn convert_id_to_mac(id: &i64) -> String {
    let mac_id_bytes = id.to_be_bytes();
    let mut mac_bytes: [u8; 6] = [0; 6];
    mac_bytes.copy_from_slice(&mac_id_bytes[2..]);
    eui48::MacAddress::new(mac_bytes).to_string(eui48::MacAddressFormat::Canonical)
}

#[derive(Serialize)]
struct GroupWithNode {
    #[serde(flatten)]
    group: node_group::Model,
    node: Vec<NodeWithMac>,
}

pub async fn get_nodes(state: web::Data<AppState>) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let result: Vec<GroupWithNode> = node_group::Entity::find()
        .find_with_related(entity::prelude::Node)
        .all(db)
        .await
        .map_err(ErrorInternalServerError)?
        .into_iter()
        .map(|(group, nodes)| {
            let nodes = nodes
                .into_iter()
                .map(|n| n.into())
                .collect::<Vec<NodeWithMac>>();
            GroupWithNode { group, node: nodes }
        })
        .collect::<Vec<_>>();

    Ok(web::Json(result))
}

pub async fn get_data(state: web::Data<AppState>) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let nodes = node::Entity::find()
        .all(db)
        .await
        .map_err(ErrorInternalServerError)?;

    let now = Utc::now();
    let one_hour_ago = now - chrono::Duration::hours(1);

    let mut result: Vec<NodeWithSensorData> = Vec::new();
    for node in nodes {
        let sensor_data = sensor_data::Entity::find()
            .filter(sensor_data::Column::Id.eq(node.id))
            .filter(sensor_data::Column::Timestamp.gt(one_hour_ago))
            .all(db)
            .await
            .map_err(ErrorInternalServerError)?;

        result.push(NodeWithSensorData {
            node: node.into(),
            sensor_data,
        });
    }

    Ok(web::Json(result))
}

pub async fn create_group(
    state: web::Data<AppState>,
    group: web::Json<CreateGroupWithoutId>,
) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let group = group.into_inner();

    let group = entity::node_group::ActiveModel {
        id: ActiveValue::NotSet,
        name: ActiveValue::Set(group.name),
    };

    let result = group.insert(db).await.map_err(ErrorInternalServerError)?;

    Ok(web::Json(result))
}

pub async fn create_node(
    state: web::Data<AppState>,
    node_request: web::Json<NodeWithMac>,
) -> actix_web::Result<impl Responder> {
    let db = &state.db;

    let node: NodeWithMac = node_request.into_inner();

    println!("Checking group ID: {:?}", node.group);

    let group_exists = entity::node_group::Entity::find_by_id(node.group)
        .one(db)
        .await
        .map_err(ErrorInternalServerError)?
        .is_some();

    if !group_exists {
        return Err(ErrorInternalServerError("Group ID does not exist"));
    }

    let node: node::Model = node
        .try_into()
        .map_err(|_| ErrorBadRequest("Invalid Mac Address"))?;
    let node = node.into_active_model();

    let result = node.insert(db).await.map_err(ErrorInternalServerError)?;

    Ok(web::Json(result))
}

pub async fn update_node(
    state: web::Data<AppState>,
    node: web::Json<UpdateNode>,
    path: Path<String>,
) -> actix_web::Result<impl Responder> {
    let db = &state.db;
    let node = node.into_inner();
    let id = path.into_inner();

    let mac_id = convert_mac_to_id(&id).map_err(ErrorBadRequest)?;
    let node = entity::node::ActiveModel {
        id: ActiveValue::Unchanged(mac_id),
        battery_minimum: ActiveValue::Set(node.battery_minimum),
        battery_maximum: ActiveValue::Set(node.battery_maximum),
        coord_la: ActiveValue::Set(node.coord_la),
        coord_lo: ActiveValue::Set(node.coord_lo),
        group: ActiveValue::Set(node.group),
    };

    let result = node.update(db).await.map_err(ErrorInternalServerError)?;

    Ok(web::Json(result))
}

/*
pub async fn delete_node(
    state: web::Data<AppState>,
    path: web::Path<Uuid>,
) -> actix_web::Result<impl Responder> {
    let id = path.into_inner();

    let db = &state.db;

    entity::node::Entity::delete_by_id(id)
        .exec(db)
        .await
        .map_err(ErrorInternalServerError)?;

    Ok(HttpResponse::Ok().finish())
}
*/
